using System;
using System.Collections.Generic;
using System.Text;

using XNAF = Microsoft.Xna.Framework;

using XGAE.Framework.Extensions;

namespace XGAE.Framework.Body.Motion
{
    public interface IParameters
    {
        /// <summary>
        /// The strength parameter is used to scale the movement actions. Zero strength means no
        /// movement or turning can occur, while a strength of one indicates maximal strength.
        /// </summary>
        /// <param name="strength"></param>
        void SetStrength(float strength);

        /// <summary>
        /// The accuracy parameter controls the precision of the actions. An accuracy of zero means
        /// the actions are completely random, while full accuracy of value one means the exact
        /// action is executed.
        /// </summary>
        /// <param name="accuracy"></param>
        void SetAccuracy(float accuracy);
    }

    /// <summary>
    /// Set of actions that allow full control of the animat's movement in 3D space. The animat's
    /// state can be characterized by a position and orientation. All the actions are specified
    /// relatively to this state. This interface does not guarantee perfect accuracy, as
    /// biologically plausible errors may be enforced before the commands are executed. This
    /// depends entirely upon the embodiment parameters given to the Motion module itself.
    /// </summary>
    public interface IRelative
    {
        /// <summary>
        /// Move in a given direction relatively to current position. This function works in full 3
        /// dimensions. The x component of the parameter vector is forwards, the y component is
        /// left, and Z corresponds to upwards, which is only applicable in certain medium where
        /// the animat can move along that dimension (for example, water). Values of magnitude 1
        /// correspond to maximum running speed, while values of 0.5 mean walking. A zero value
        /// force the movement to stop along that dimension.
        /// </summary>
        void Step(XNAF.Vector3 r_direction);

        /// <summary>
        /// Rotate the view by a given set of angles relatively to the current orientation. The
        /// parameter is a set of three angles corresponding to pitch, yaw and roll respectively.
        /// Each angle is specified in degrees, best between -180 and +180. The roll is not
        /// guaranteed to be taken into account. The user should expect the pitch to be clamped
        /// once it reaches vertical positions looking either straight up and down.
        /// </summary>
        void Turn(Angles r_angles);

        /// <summary>
        /// Rotate the view by a certain number of degrees relatively to the current orientation.
        /// This action behaves in the same way as the more generic Turn() function, except the
        /// parameters are unrolled and reduced to two, with default values provided.
        /// </summary>
        void Turn(float yaw, float pitch);

        /// <summary>
        /// Jump upwards, combine with Step() to leap in a specific direction.
        /// </summary>
        void Jump();

        /// <summary>
        /// Duck down, use Step() to crawl in a specific direction.
        /// </summary>
        void Duck();
    }
}