using System;
using System.Collections.Generic;
using System.Text;

using XNAF = Microsoft.Xna.Framework;

using XGAE.Framework;
using XGAE.Framework.Extensions;
using xBody = XGAE.Framework.Body;

namespace XGAE.Body.Motion
{
    /// <summary>
    /// Base implementation for all the interfaces providing inline wrappers and convenient 
    /// access to the appropriate variables in the module.
    /// </summary>
    interface IMotionInterface
    {
        /// <summary>
        /// Reference to the owner component
        /// </summary>
        MotionModule Self { get; }

        /// <summary>
        /// Wrappers to retrieve the backend from the module. 
        /// </summary>
        /// <returns></returns>
        Backend.IMotion GetBackend();
    }

    /// <summary>
    /// Set of actions that allow full control of the animat's movement in 3D space. The animat's
    /// state can be characterized by a position and orientation. All the actions are specified
    /// relatively to this state. This interface does not guarantee perfect accuracy, as
    /// biologically plausible errors may be enforced before the commands are executed. This
    /// depends entirely upon the embodiment parameters given to the Motion module itself.
    /// </summary>
    class MotionRelative : Interface, xBody.Motion.IRelative, IMotionInterface
    {
        private MotionModule _self;

        public MotionModule Self
        {
            get
            {
                return this._self;
            }
        }

        private Backend.IMotion _backend;

        public Backend.IMotion Backend
        {
            get { return this._backend; }
        }

        #region C'tors

        /// <summary>
        /// Constructor called by the module when it creates a new interface.
        /// </summary>
        /// <param name="?"></param>
	    public MotionRelative(MotionModule m)
	    :	base("Relative", 0x0004)		    
	    {
            this._self = m;
	    }

        #endregion

        public Backend.IMotion GetBackend()
        {
            return this.Self.Backend;
        }

        /// <summary>
        /// Move in a given direction relatively to current position. This function works in full 3
        /// dimensions. The x component of the parameter vector is forwards, the y component is
        /// left, and Z corresponds to upwards, which is only applicable in certain medium where
        /// the animat can move along that dimension (for example, water). Values of magnitude 1
        /// correspond to maximum running speed, while values of 0.5 mean walking. A zero value
        /// force the movement to stop along that dimension.
        /// </summary>
        /// <param name="r_direction"></param>
        public void Step(XNAF.Vector3 r_direction)
        {
            this.GetBackend().Step(r_direction);
        }

        /// <summary>
        /// Rotate the view by a given set of angles relatively to the current orientation. The
        /// parameter is a set of three angles corresponding to pitch, yaw and roll respectively.
        /// Each angle is specified in degrees, best between -180 and +180. The roll is not
        /// guaranteed to be taken into account. The user should expect the pitch to be clamped
        /// once it reaches vertical positions looking either straight up and down.
        /// </summary>
        /// <param name="r_angles"></param>
        public void Turn(Angles r_angles)
        {
            this.GetBackend().Turn(r_angles);
        }

        public void Turn(float yaw, float pitch)
        {
	        this.GetBackend().Turn( new Angles( pitch, yaw, 0.0f) );
        }

        /// <summary>
        /// Jump upwards, combine with Step() to leap in a specific direction.
        /// </summary>
        public void Jump()
        {
            this.GetBackend().Jump();
        }

        /// <summary>
        /// /// Duck down, use move to crawl in a specific direction.
        /// </summary>
        public void Duck()
        {
	        this.GetBackend().Duck();
        }
    }

    class MotionParameters : Interface, xBody.Motion.IParameters, IMotionInterface
    {
        #region IMotionInterface Implementation

        private MotionModule _self;

        public MotionModule Self
        {
            get
            {
                return this._self;
            }
        }

        public Backend.IMotion GetBackend()
        {
            return this.Self.Backend;
        }

        #endregion

        #region C'tors

        /// <summary>
        /// Constructor called by the module when it creates a new interface.
        /// </summary>
        /// <param name="m"></param>
	    public MotionParameters(MotionModule m)
	    :	base("Parameters", 0x0001)		
	    {
            this._self = m;
	    }

        #endregion

        #region IParameters Implementations

        /// <summary>
        /// The strength parameter is used to scale the movement actions. Zero strength means no
        /// movement or turning can occur, while a strength of one indicates maximal strength.
        /// </summary>
        /// <param name="strength"></param>
        public void SetStrength(float strength)
        {
            this.Self.Strength = strength;
        }

        /// <summary>
        /// The accuracy parameter controls the precision of the actions. An accuracy of zero means
        /// the actions are completely random, while full accuracy of value one means the exact
        /// action is executed.
        /// </summary>
        /// <param name="accuracy"></param>
        public void SetAccuracy(float accuracy)
        {
            this.Self.Accuracy = accuracy;
        }

        #endregion
    }
}
